import pytest

from wr_attrs import Attr, Attrs, container


def test_attrs_cls_names_is_a_property():
    assert isinstance(Attrs._names_, property)


def test_attrs_cls_all_is_a_property():
    assert isinstance(Attrs._all_, property)


def test_attrs_cls_tagged_is_a_property():
    assert callable(Attrs._tagged_)


def test_cannot_access_non_attr_via_attrs():
    @container
    class C:
        x = Attr()
        y = 10  # non-attr

    c = C()

    with pytest.raises(AttributeError) as exc_info:
        _ = C.attrs.y  # noqa
    assert 'C.y is not an Attr' in str(exc_info.value)

    with pytest.raises(AttributeError) as exc_info:
        _ = c.attrs.y  # noqa
    assert 'C.y is not an Attr' in str(exc_info.value)


def test_cannot_reset_attr_on_attrs(xy_container_cls):
    with pytest.raises(AttributeError):
        xy_container_cls.attrs.x = Attr('x', default=100)


def test_cannot_reset_attrs(xy_container_cls):
    c = xy_container_cls()

    with pytest.raises(AttributeError) as exc_info:
        c.attrs = Attrs(c)

    assert 'C.attrs is read-only' in str(exc_info.value)



from wr_attrs import Attr, container


def test_init_value_decorator():
    @container
    class C:
        @Attr.init_value
        def x(self, attr):
            attr.value = 5

        y = Attr()

        @y.init_value
        def y(self, attr):
            attr.value = 20

    c = C()

    assert c.x == 5

    c.x = 10
    assert c.x == 10

    assert c.y == 20


def test_get_value_decorator():
    @container
    class C:
        @Attr.get_value
        def x(self, attr):
            return attr.value * 5 if attr.value else attr.value

        y = Attr()

        @y.get_value
        def y(self, attr):
            return attr.value * 2 if attr.value else attr.value

    c = C()

    assert c.x is None

    c.x = 10
    assert c.x == 50

    c.y = 10
    assert c.y == 20


def test_set_value_decorator():
    @container
    class C:
        @Attr.set_value
        def x(self, attr, value):
            attr.value = value * 5

        y = Attr()

        @y.set_value
        def y(self, attr, value):
            attr.value = value * 2

    c = C()

    assert c.x is None
    c.x = 5
    assert c.x == 25

    c.y = 5
    assert c.y == 10



import pytest


def test_repr(xy_container_cls):
    c = xy_container_cls()
    assert '<BoundAttr C.x>' == repr(c.attrs.x)


def test_class_bound_attr_has_no_value(xy_container_cls):
    with pytest.raises(TypeError):
        _ = xy_container_cls.attrs.x.value  # noqa



import pytest

from wr_attrs import Attr


def test_repr():
    x = Attr(name='x')
    assert repr(x) == "<Attr 'x'>"


def test_cannot_set_nonexistent_attribute_of_attr():
    x = Attr()

    with pytest.raises(AttributeError):
        x.safe = True


def test_cannot_set_default_value_and_required_together():
    with pytest.raises(ValueError):
        Attr(required=True, default=1)
